# This file was auto-generated by Fern from our API Definition.

import typing
from ..core.client_wrapper import SyncClientWrapper
from .plot.client import PlotClient
from .equation.client import EquationClient
from .. import core
from ..core.request_options import RequestOptions
from ..types.extract_text_response import ExtractTextResponse
from ..core.pydantic_utilities import parse_obj_as
from ..errors.unprocessable_entity_error import UnprocessableEntityError
from ..types.http_validation_error import HttpValidationError
from json.decoder import JSONDecodeError
from ..core.api_error import ApiError
from ..types.parse_response import ParseResponse
from ..types.extract_constants_response import ExtractConstantsResponse
from ..core.client_wrapper import AsyncClientWrapper
from .plot.client import AsyncPlotClient
from .equation.client import AsyncEquationClient

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class DocumentClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper
        self.plot = PlotClient(client_wrapper=self._client_wrapper)
        self.equation = EquationClient(client_wrapper=self._client_wrapper)

    def text(
        self,
        *,
        file: core.File,
        method: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ExtractTextResponse:
        """
        Extracts text from documents

        Parameters
        ----------
        file : core.File
            See core.File for more documentation

        method : typing.Optional[str]
            Method to use for text-only extraction.It uses a very simple pdf text extractor.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ExtractTextResponse
            Successful Response

        Examples
        --------
        from axiomatic import Axiomatic

        client = Axiomatic(
            api_key="YOUR_API_KEY",
        )
        client.document.text()
        """
        _response = self._client_wrapper.httpx_client.request(
            "document/text",
            method="POST",
            params={
                "method": method,
            },
            data={},
            files={
                "file": file,
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    ExtractTextResponse,
                    parse_obj_as(
                        type_=ExtractTextResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    def parse(
        self,
        *,
        file: core.File,
        method: typing.Optional[str] = None,
        ocr: typing.Optional[bool] = None,
        layout_model: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ParseResponse:
        """
        Extracts text from documents. It uses advanced pdf segmentation.

        Parameters
        ----------
        file : core.File
            See core.File for more documentation

        method : typing.Optional[str]
            Method to use for text extraction

        ocr : typing.Optional[bool]
            Whether to use OCR

        layout_model : typing.Optional[str]
            Method for layout parsing

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ParseResponse
            Successful Response

        Examples
        --------
        from axiomatic import Axiomatic

        client = Axiomatic(
            api_key="YOUR_API_KEY",
        )
        client.document.parse()
        """
        _response = self._client_wrapper.httpx_client.request(
            "document/parse",
            method="POST",
            params={
                "method": method,
                "ocr": ocr,
                "layout_model": layout_model,
            },
            data={},
            files={
                "file": file,
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    ParseResponse,
                    parse_obj_as(
                        type_=ParseResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    def constants(
        self, *, file: core.File, constants: typing.List[str], request_options: typing.Optional[RequestOptions] = None
    ) -> ExtractConstantsResponse:
        """
        Extracts specific constants from documents

        Parameters
        ----------
        file : core.File
            See core.File for more documentation

        constants : typing.List[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ExtractConstantsResponse
            Successful Response

        Examples
        --------
        from axiomatic import Axiomatic

        client = Axiomatic(
            api_key="YOUR_API_KEY",
        )
        client.document.constants(
            constants=["constants"],
        )
        """
        _response = self._client_wrapper.httpx_client.request(
            "document/constants",
            method="POST",
            data={
                "constants": constants,
            },
            files={
                "file": file,
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    ExtractConstantsResponse,
                    parse_obj_as(
                        type_=ExtractConstantsResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)


class AsyncDocumentClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper
        self.plot = AsyncPlotClient(client_wrapper=self._client_wrapper)
        self.equation = AsyncEquationClient(client_wrapper=self._client_wrapper)

    async def text(
        self,
        *,
        file: core.File,
        method: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ExtractTextResponse:
        """
        Extracts text from documents

        Parameters
        ----------
        file : core.File
            See core.File for more documentation

        method : typing.Optional[str]
            Method to use for text-only extraction.It uses a very simple pdf text extractor.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ExtractTextResponse
            Successful Response

        Examples
        --------
        import asyncio

        from axiomatic import AsyncAxiomatic

        client = AsyncAxiomatic(
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.document.text()


        asyncio.run(main())
        """
        _response = await self._client_wrapper.httpx_client.request(
            "document/text",
            method="POST",
            params={
                "method": method,
            },
            data={},
            files={
                "file": file,
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    ExtractTextResponse,
                    parse_obj_as(
                        type_=ExtractTextResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    async def parse(
        self,
        *,
        file: core.File,
        method: typing.Optional[str] = None,
        ocr: typing.Optional[bool] = None,
        layout_model: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ParseResponse:
        """
        Extracts text from documents. It uses advanced pdf segmentation.

        Parameters
        ----------
        file : core.File
            See core.File for more documentation

        method : typing.Optional[str]
            Method to use for text extraction

        ocr : typing.Optional[bool]
            Whether to use OCR

        layout_model : typing.Optional[str]
            Method for layout parsing

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ParseResponse
            Successful Response

        Examples
        --------
        import asyncio

        from axiomatic import AsyncAxiomatic

        client = AsyncAxiomatic(
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.document.parse()


        asyncio.run(main())
        """
        _response = await self._client_wrapper.httpx_client.request(
            "document/parse",
            method="POST",
            params={
                "method": method,
                "ocr": ocr,
                "layout_model": layout_model,
            },
            data={},
            files={
                "file": file,
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    ParseResponse,
                    parse_obj_as(
                        type_=ParseResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    async def constants(
        self, *, file: core.File, constants: typing.List[str], request_options: typing.Optional[RequestOptions] = None
    ) -> ExtractConstantsResponse:
        """
        Extracts specific constants from documents

        Parameters
        ----------
        file : core.File
            See core.File for more documentation

        constants : typing.List[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ExtractConstantsResponse
            Successful Response

        Examples
        --------
        import asyncio

        from axiomatic import AsyncAxiomatic

        client = AsyncAxiomatic(
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.document.constants(
                constants=["constants"],
            )


        asyncio.run(main())
        """
        _response = await self._client_wrapper.httpx_client.request(
            "document/constants",
            method="POST",
            data={
                "constants": constants,
            },
            files={
                "file": file,
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    ExtractConstantsResponse,
                    parse_obj_as(
                        type_=ExtractConstantsResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    typing.cast(
                        HttpValidationError,
                        parse_obj_as(
                            type_=HttpValidationError,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
from .statement_type import StatementType
import pydantic
from .statement_validation import StatementValidation
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class UnformalizableStatement(UniversalBaseModel):
    type: typing.Optional[StatementType] = None
    text: str = pydantic.Field()
    """
    The natural language content of the statement.
    """

    formalization: typing.Optional[str] = None
    validation: typing.Optional[StatementValidation] = pydantic.Field(default=None)
    """
    The validation result of the statement.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
from .cost_function import CostFunction
from .structure_constraint import StructureConstraint
from .parameter_constraint import ParameterConstraint
from .unformalizable_statement import UnformalizableStatement
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import pydantic


class StatementDictionary(UniversalBaseModel):
    """
    A dictionary of statements by type.
    """

    cost_functions: typing.Optional[typing.List[CostFunction]] = None
    structure_constraints: typing.Optional[typing.List[StructureConstraint]] = None
    parameter_constraints: typing.Optional[typing.List[ParameterConstraint]] = None
    unformalizable_statements: typing.Optional[typing.List[UnformalizableStatement]] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
import pydantic
from .computation import Computation
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class Z3Expression(UniversalBaseModel):
    """
    The model for a Z3 expression.
    The validation is such that an instance contains a code that can be parsed into a z3 expression and, if mapping is set, it contains a valid Computation for every variable in the expression.
    It does not check if the expression can be evaluated, this depends on the target Netlist.
    """

    type: typing.Optional[typing.Literal["Z3_EXPRESSION"]] = None
    code: str = pydantic.Field()
    """
    A string with the z3 expression.
    """

    default_tolerance: typing.Optional[float] = pydantic.Field(default=None)
    """
    The tolerance to be used when comparing the expression with a value.
    """

    mapping: typing.Optional[typing.Dict[str, typing.Optional[Computation]]] = pydantic.Field(default=None)
    """
    A mapping of the variables used in the expression to Computations.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

from .axes_info import AxesInfo
from .bundle import Bundle
from .bundle_settings_value import BundleSettingsValue
from .computation import Computation
from .computation_arguments_value import ComputationArgumentsValue
from .contour_cv_2 import ContourCv2
from .cost_function import CostFunction
from .create_api_key_response import CreateApiKeyResponse
from .create_api_keyrequest import CreateApiKeyrequest
from .create_user_request import CreateUserRequest
from .create_user_response import CreateUserResponse
from .dict_item import DictItem
from .equation_extraction import EquationExtraction
from .equation_extraction_response import EquationExtractionResponse
from .equation_processing_response import EquationProcessingResponse
from .equation_validation_result import EquationValidationResult
from .execute_code_response import ExecuteCodeResponse
from .extract_constants_response import ExtractConstantsResponse
from .extract_text_response import ExtractTextResponse
from .extracted_point import ExtractedPoint
from .extracted_points import ExtractedPoints
from .find_mapping_response import FindMappingResponse
from .find_user_response import FindUserResponse
from .formalize_circuit_response import FormalizeCircuitResponse
from .formalize_response import FormalizeResponse
from .generate_code_response import GenerateCodeResponse
from .generate_component_code_response import GenerateComponentCodeResponse
from .generate_lens_code_response import GenerateLensCodeResponse
from .get_optimizable_parameters_response import GetOptimizableParametersResponse
from .get_spectrum_response import GetSpectrumResponse
from .get_spectrum_response_spectrum_value_item import GetSpectrumResponseSpectrumValueItem
from .http_validation_error import HttpValidationError
from .informalize_statement_response import InformalizeStatementResponse
from .net import Net
from .netlist import Netlist
from .optimization_history import OptimizationHistory
from .optimize_config import OptimizeConfig
from .optimize_netlist_response import OptimizeNetlistResponse
from .optimize_placement_body_response import OptimizePlacementBodyResponse
from .parameter import Parameter
from .parameter_constraint import ParameterConstraint
from .parameter_output_lower_bound import ParameterOutputLowerBound
from .parameter_output_upper_bound import ParameterOutputUpperBound
from .parse_response import ParseResponse
from .parse_statement_response import ParseStatementResponse
from .pdk_type import PdkType
from .pic_instance import PicInstance
from .pic_instance_info_value import PicInstanceInfoValue
from .pic_instance_settings_value import PicInstanceSettingsValue
from .pic_warning import PicWarning
from .placement import Placement
from .plot_info import PlotInfo
from .plot_parser_output import PlotParserOutput
from .refine_code_response import RefineCodeResponse
from .refine_component_code_response import RefineComponentCodeResponse
from .schedule_job_response import ScheduleJobResponse
from .solution_response import SolutionResponse
from .solution_response_solution_value import SolutionResponseSolutionValue
from .statement import Statement
from .statement_dictionary import StatementDictionary
from .statement_type import StatementType
from .statement_validation import StatementValidation
from .statement_validation_dictionary import StatementValidationDictionary
from .status_response import StatusResponse
from .structure_constraint import StructureConstraint
from .structure_function_call import StructureFunctionCall
from .structure_function_call_arguments_value import StructureFunctionCallArgumentsValue
from .structure_function_call_expected_result import StructureFunctionCallExpectedResult
from .summarizer_response import SummarizerResponse
from .tools_list_response import ToolsListResponse
from .unformalizable_statement import UnformalizableStatement
from .user_requirement import UserRequirement
from .validate_netlist_response import ValidateNetlistResponse
from .validate_response import ValidateResponse
from .validation_error import ValidationError
from .validation_error_loc_item import ValidationErrorLocItem
from .variable_requirement import VariableRequirement
from .verify_circuit_code_response import VerifyCircuitCodeResponse
from .verify_response import VerifyResponse
from .z_3_expression import Z3Expression

__all__ = [
    "AxesInfo",
    "Bundle",
    "BundleSettingsValue",
    "Computation",
    "ComputationArgumentsValue",
    "ContourCv2",
    "CostFunction",
    "CreateApiKeyResponse",
    "CreateApiKeyrequest",
    "CreateUserRequest",
    "CreateUserResponse",
    "DictItem",
    "EquationExtraction",
    "EquationExtractionResponse",
    "EquationProcessingResponse",
    "EquationValidationResult",
    "ExecuteCodeResponse",
    "ExtractConstantsResponse",
    "ExtractTextResponse",
    "ExtractedPoint",
    "ExtractedPoints",
    "FindMappingResponse",
    "FindUserResponse",
    "FormalizeCircuitResponse",
    "FormalizeResponse",
    "GenerateCodeResponse",
    "GenerateComponentCodeResponse",
    "GenerateLensCodeResponse",
    "GetOptimizableParametersResponse",
    "GetSpectrumResponse",
    "GetSpectrumResponseSpectrumValueItem",
    "HttpValidationError",
    "InformalizeStatementResponse",
    "Net",
    "Netlist",
    "OptimizationHistory",
    "OptimizeConfig",
    "OptimizeNetlistResponse",
    "OptimizePlacementBodyResponse",
    "Parameter",
    "ParameterConstraint",
    "ParameterOutputLowerBound",
    "ParameterOutputUpperBound",
    "ParseResponse",
    "ParseStatementResponse",
    "PdkType",
    "PicInstance",
    "PicInstanceInfoValue",
    "PicInstanceSettingsValue",
    "PicWarning",
    "Placement",
    "PlotInfo",
    "PlotParserOutput",
    "RefineCodeResponse",
    "RefineComponentCodeResponse",
    "ScheduleJobResponse",
    "SolutionResponse",
    "SolutionResponseSolutionValue",
    "Statement",
    "StatementDictionary",
    "StatementType",
    "StatementValidation",
    "StatementValidationDictionary",
    "StatusResponse",
    "StructureConstraint",
    "StructureFunctionCall",
    "StructureFunctionCallArgumentsValue",
    "StructureFunctionCallExpectedResult",
    "SummarizerResponse",
    "ToolsListResponse",
    "UnformalizableStatement",
    "UserRequirement",
    "ValidateNetlistResponse",
    "ValidateResponse",
    "ValidationError",
    "ValidationErrorLocItem",
    "VariableRequirement",
    "VerifyCircuitCodeResponse",
    "VerifyResponse",
    "Z3Expression",
]



from .relation_graph import generate_relation_graph
from .. import EquationExtractionResponse
import os
import re

HTML_HEAD = """
    <!DOCTYPE html>
    <html>
    <head>
        <title>Equation Extraction Report</title>
        <meta charset="UTF-8">
        <script src="https://polyfill.io/v3/polyfill.min.js?features=es6"></script>
        <script id="MathJax-script" src="https://cdn.jsdelivr.net/npm/mathjax@3/es5/tex-mml-chtml.js"></script>
        <style>
            :root {
                --system1-bg: #fff;
                --system1-text: #000;
                --system1-border: #000;
                --system1-highlight: #000;
            }
            
            body { 
                font-family: 'Chicago', 'Helvetica', sans-serif;
                margin: 8px; 
                padding: 0;
                background-color: var(--system1-bg);
                color: var(--system1-text);
                font-size: 12px;
            }
            
            /* Menu Bar Styles */
            .menu-bar {
                background: var(--system1-bg);
                border-bottom: 1px solid var(--system1-border);
                padding: 2px;
                display: flex;
                gap: 16px;
                margin-bottom: 8px;
            }
            
            .menu-button {
                background: var(--system1-bg);
                border: 1.5px solid var(--system1-border);
                border-radius: 8px;
                padding: 2px 8px;
                font-family: 'Chicago', 'Helvetica', sans-serif;
                font-size: 12px;
                cursor: pointer;
                position: relative;
                white-space: nowrap;
            }
            
            .menu-button:active {
                background: var(--system1-text);
                color: var(--system1-bg);
            }
            
            h1 { 
                margin: 0 0 16px 0;
                padding: 4px;
                font-size: 14px;
                font-weight: bold;
                text-align: center;
                border-bottom: 1.5px solid var(--system1-border);
            }
            
            .container {
                display: flex;
                gap: 16px;
                height: calc(100vh - 100px);
            }
            
            .equation-list {
                width: 300px;
                border: 1.5px solid var(--system1-border);
                border-radius: 8px;
                padding: 8px;
                overflow-y: auto;
            }
            
            .equation-item { 
                border: 1.5px solid transparent;
                margin-bottom: 12px;
                padding: 8px;
                cursor: pointer;
                border-radius: 4px;
            }
            
            .equation-item:hover { 
                background: var(--system1-text);
                color: var(--system1-bg);
            }
            
            .equation-item.active {
                background: var(--system1-text);
                color: var(--system1-bg);
            }
            
            .details-panel {
                flex: 1;
                border: 1.5px solid var(--system1-border);
                border-radius: 8px;
                padding: 16px;
                overflow-y: auto;
            }
            
            .details-panel.empty {
                display: flex;
                align-items: center;
                justify-content: center;
                font-style: italic;
            }
            
            .equation-details h2 {
                margin: 0 0 16px 0;
                font-size: 14px;
                text-align: center;
            }
            
            .equation-details p {
                margin: 16px 0 8px 0;
            }
            
            .symbol-table { 
                width: 100%; 
                border-collapse: collapse;
                margin-top: 8px;
                border: 1.5px solid var(--system1-border);
            }
            
            .symbol-table td, .symbol-table th { 
                border: 1px solid var(--system1-border);
                padding: 6px 8px;
                text-align: left;
            }
            
            .symbol-table th {
                background: var(--system1-text);
                color: var(--system1-bg);
                font-weight: bold;
            }
            
            .equation-title {
                margin: 0 0 8px 0;
                font-size: 12px;
                font-weight: bold;
            }

            /* MathJax container spacing */
            .mjx-chtml {
                margin: 8px 0 !important;
            }

            /* Classic System 1 scrollbar */
            ::-webkit-scrollbar {
                width: 16px;
                height: 16px;
            }
            
            ::-webkit-scrollbar-track {
                background: var(--system1-bg);
                border: 1.5px solid var(--system1-border);
            }
            
            ::-webkit-scrollbar-thumb {
                background: var(--system1-text);
                border: 1.5px solid var(--system1-border);
            }
            
            ::-webkit-scrollbar-button {
                display: none;
            }
        </style>
        <script>
            function showDetails(eqId) {
                document.querySelectorAll('.equation-item').forEach(item => {
                    item.classList.remove('active');
                });
                
                document.getElementById('eq-item-' + eqId).classList.add('active');
                
                document.querySelectorAll('.equation-details').forEach(detail => {
                    detail.style.display = 'none';
                });
                document.getElementById('empty-message').style.display = 'none';
                document.getElementById('details-' + eqId).style.display = 'block';
            }
        </script>
    </head>
"""


def create_report(report_data: EquationExtractionResponse, report_path: str = "./report.html"):
    """
    Creates an HTML report for the extracted equations.
    """
    html_content = (
        HTML_HEAD
        + """
    <body>
        <div class="menu-bar">
            <a href="#" onclick="showHypergraph()">Relation Graph</a>
        </div>
        
        <h1>Equation Extraction Report</h1>
        <div class="container">
            <div class="equation-list">
    """
    )

    # Add equations to the list
    for i, eq in enumerate(report_data.equations):
        html_content += f"""
                <div class="equation-item" id="eq-item-{i}" onclick="showDetails({i})">
                    <div class="equation-title">{eq.name}</div>
                    \\[{eq.original_format}\\]
                </div>
        """

    html_content += """
            </div>
            <div class="details-panel">
                <div id="empty-message" style="text-align: center; color: #666;">
                    Select an equation to view details
                </div>
    """

    # Add equation details
    for i, eq in enumerate(report_data.equations):
        html_content += f"""
                <div class="equation-details" id="details-{i}" style="display: none;">
                    <h2>{eq.name}</h2>
                    <p><strong>Description:</strong></p>
                    <p>{eq.description}</p>
                    
                    <p><strong>Original Format:</strong></p>
                    \\[{eq.original_format}\\]
                    
                    <p><strong>Symbols:</strong></p>
                    <table class="symbol-table">
                        <tr>
                            <th>Symbol</th>
                            <th>Description</th>
                        </tr>
        """

        for symbol in eq.latex_symbols:
            html_content += f"""
                        <tr>
                            <td>\\({symbol.key}\\)</td>
                            <td>{symbol.key}</td>
                        </tr>
            """

        html_content += """
                    </table>
                </div>
        """

    html_content += """
            </div>
        </div>
        
        <!-- Graph Modal -->
        <div id="hypergraph-modal" style="display: none; position: fixed; top: 0; left: 0; width: 100%; height: 100%; 
             background: rgba(0,0,0,0.8); z-index: 1000;">
            <div style="position: relative; width: 90%; height: 90%; margin: 2% auto; background: var(--system1-bg); 
                 padding: 20px; border-radius: 8px; overflow: hidden;">
                <button onclick="hideHypergraph()" style="position: absolute; right: 10px; top: 10px; z-index: 1001;" 
                        class="menu-button">Close</button>
                <div id="graph-container" style="width: 100%; height: 100%;">
    """

    # Generate the relation graph as a separate HTML file
    graph_html = generate_relation_graph(report_data.equations)
    graph_file_path = os.path.join(os.path.dirname(report_path), "relation_graph.html")
    with open(graph_file_path, "w", encoding="utf-8") as f:
        f.write(graph_html)

    # Extract just the graph content from the full HTML
    # This avoids duplicate HTML, head, body tags
    graph_content = extract_graph_content(graph_html)
    html_content += graph_content

    # Add the relation graph link to the main menu with proper styling
    html_content = html_content.replace(
        '<a href="#" onclick="showHypergraph()">Relation Graph</a>',
        f'<a href="{os.path.basename(graph_file_path)}" target="_blank" class="menu-button">Relation Graph</a>',
    )

    # Remove the menu-item related code since we're using menu-button class
    html_content = re.sub(r"\.menu-item \{.*?\}", "", html_content, flags=re.DOTALL)

    # Remove the showHypergraph and hideHypergraph functions
    html_content = re.sub(
        r"<script>\s*function showHypergraph\(\).*?function hideHypergraph\(\).*?</script>",
        "",
        html_content,
        flags=re.DOTALL,
    )

    # Close all remaining tags if needed
    if "</div></div></div>" not in html_content:
        html_content += """
                </div>
            </div>
        </div>
        """

    with open(report_path, "w", encoding="utf-8") as f:
        f.write(html_content)


def extract_graph_content(graph_html: str) -> str:
    """
    Extracts the relevant graph content from the full HTML generated by pyvis.

    Args:
        graph_html: Full HTML string from pyvis

    Returns:
        String containing just the graph-specific content
    """
    # Find where the actual graph content starts (after the body tag)
    body_start = graph_html.find("<body>")
    if body_start != -1:
        content_start = body_start + len("<body>")
        # Find where the content ends (before closing body tag)
        body_end = graph_html.find("</body>")
        if body_end != -1:
            # Extract just the graph content
            return graph_html[content_start:body_end]

    # Fallback if we can't parse it properly
    return graph_html



from IPython import get_ipython  # type: ignore
from IPython.display import HTML, display  # type: ignore

import platformdirs  # type: ignore
import os
import sys
import time
from . import Axiomatic


class AXMagic:
    """Class implementing magic functions for IPython.
    Import with `%load_ext axiomatic.magic`."""

    client: Axiomatic
    query: str

    def __init__(self):
        self.folder = platformdirs.user_config_dir("axiomatic")

        self.client = Axiomatic()
        self.query = ""

    def ax_api(self, *args, **kwargs):
        pass

    def axquery(self, query, cell=None):
        if cell:
            # REFINE
            try:
                exec(cell)
                feedback = "Code executed successfully."
            except Exception as e:
                feedback = f"Errors:\n{e}"
            print(feedback)
            current_query = f"{self.query}\n{query}"
            result = self.client.pic.circuit.refine(query=current_query, code=cell, feedback=feedback)
        else:
            # GENERATE FROM SCRATCH
            self.query = query
            result = self.client.pic.circuit.generate(query=query)

        # Process output
        pre_thought = result.raw_content.split("<thought>")[0]
        thought = result.thought_text.replace("\n", "<br>")
        if not thought:
            output = result.raw_content.replace("\n", "<br>")
        else:
            output = pre_thought + "<br>" + thought
        html_content = f"""<div style='font-family: Arial, sans-serif; line-height: 1.5;'>"
<div style='color: #6EB700;'><strong>AX:</strong> {output}</div>"""
        display(HTML(html_content))

        # Process code
        # remove last three lines (saving file)
        if result.code:
            code = "\n".join(result.code.split("\n")[:-3] + ["c"])
            if "google.colab" in sys.modules:
                # When running in colab
                from google.colab import _frontend  # type: ignore

                _frontend.create_scratch_cell(f"""# {query}\n# %%ax_fix\n{code}""", bottom_pane=True)
            else:
                # When running in jupyter
                get_ipython().set_next_input(f"# %%ax_fix\n{code}", replace=False)

    def ax_query(self, query, cell=None):
        # Updates the target circuit query
        self.query = query
        return self.axquery(query, cell)

    def ax_fix(self, query, cell=None):
        # Runs again without updating the query
        return self.axquery(query, cell)

    def ax_tool(self, tool, cell):
        """
        A custom IPython cell magic that sends python code in a cell to the tools scheduling API
        using Axiomatic's `client.tools.schedule` and wait for the tool execution to finish and
        show the 'stdout' of the script

        Usage:

            %%tool_schedule [optional_tool_name]
            <some code here>

        Example:

            %%tool_schedule fdtd
            print("Hello from this cell!")
            x = 1 + 2

        This cell won't run as Python code; instead, the text will be sent to the tool_schedule method
        of the Axiomatic client.

        Available Tools:
            - fdtd
            - femwell
            - jaxfem
            - optiland
            - pyspice
            - sax-gdsfactory
        """
        if not tool.strip():
            print("Please provider a tool name when calling this magic like:  %%tool_schedule [optional_tool_name]")
            tools = self.client.tools.list()
            print("Available tools are:")
            for tool in tools.tools_list:
                print(f"- {tool}")
        else:
            tool_name = tool.strip()
            code_string = cell

            output = self.client.tools.schedule(
                tool_name=tool_name,
                code=code_string,
            )
            if output.is_success is True:
                job_id = output.job_id
                result = self.client.tools.status(job_id=output.job_id)
                print(f"job_id: {job_id}")
                while True:
                    result = self.client.tools.status(job_id=output.job_id)
                    if result.status == "PENDING" or result.status == "RUNNING":
                        time.sleep(3)
                    else:
                        if result.status == "SUCCEEDED":
                            os.environ["TOOL_RESULT"] = result.output
                            get_ipython().user_ns["tool_result"] = result.output
                            print(result.output)
                        else:
                            print(result.error_trace)
                        break
            else:
                print(output.error_trace)


def ax_help(value: str):
    print(
        """
Available commands:

- `%load_ext axiomatic_pic` loads the ipython extension.
- `%ax_query` returns the requested circuit using our experimental API
- `%%ax_fix` edit the given code
- `%%ax_tool tool_name` executes python code for a given tool. Available tools are:
  `fdtd, femwell, jaxfem, optiland, pyspice and sax-gdsfactory`
"""
    )


def load_ipython_extension(ipython):
    ax_magic = AXMagic()
    ipython.register_magic_function(ax_magic.ax_query, "line_cell")
    ipython.register_magic_function(ax_magic.ax_fix, "line_cell")
    ipython.register_magic_function(ax_magic.ax_api, "line")
    ipython.register_magic_function(ax_magic.ax_tool, "cell")
    ipython.register_magic_function(ax_help, "line")



# This file was auto-generated by Fern from our API Definition.

from .types import (
    AxesInfo,
    Bundle,
    BundleSettingsValue,
    Computation,
    ComputationArgumentsValue,
    ContourCv2,
    CostFunction,
    CreateApiKeyResponse,
    CreateApiKeyrequest,
    CreateUserRequest,
    CreateUserResponse,
    DictItem,
    EquationExtraction,
    EquationExtractionResponse,
    EquationProcessingResponse,
    EquationValidationResult,
    ExecuteCodeResponse,
    ExtractConstantsResponse,
    ExtractTextResponse,
    ExtractedPoint,
    ExtractedPoints,
    FindMappingResponse,
    FindUserResponse,
    FormalizeCircuitResponse,
    FormalizeResponse,
    GenerateCodeResponse,
    GenerateComponentCodeResponse,
    GenerateLensCodeResponse,
    GetOptimizableParametersResponse,
    GetSpectrumResponse,
    GetSpectrumResponseSpectrumValueItem,
    HttpValidationError,
    InformalizeStatementResponse,
    Net,
    Netlist,
    OptimizationHistory,
    OptimizeConfig,
    OptimizeNetlistResponse,
    OptimizePlacementBodyResponse,
    Parameter,
    ParameterConstraint,
    ParameterOutputLowerBound,
    ParameterOutputUpperBound,
    ParseResponse,
    ParseStatementResponse,
    PdkType,
    PicInstance,
    PicInstanceInfoValue,
    PicInstanceSettingsValue,
    PicWarning,
    Placement,
    PlotInfo,
    PlotParserOutput,
    RefineCodeResponse,
    RefineComponentCodeResponse,
    ScheduleJobResponse,
    SolutionResponse,
    SolutionResponseSolutionValue,
    Statement,
    StatementDictionary,
    StatementType,
    StatementValidation,
    StatementValidationDictionary,
    StatusResponse,
    StructureConstraint,
    StructureFunctionCall,
    StructureFunctionCallArgumentsValue,
    StructureFunctionCallExpectedResult,
    SummarizerResponse,
    ToolsListResponse,
    UnformalizableStatement,
    UserRequirement,
    ValidateNetlistResponse,
    ValidateResponse,
    ValidationError,
    ValidationErrorLocItem,
    VariableRequirement,
    VerifyCircuitCodeResponse,
    VerifyResponse,
    Z3Expression,
)
from .errors import UnprocessableEntityError
from . import code_execution, document, formalization, fso, lean, pic, requirements, tools
from .client import AsyncAxiomatic, Axiomatic
from .environment import AxiomaticEnvironment
from .version import __version__

__all__ = [
    "AsyncAxiomatic",
    "AxesInfo",
    "Axiomatic",
    "AxiomaticEnvironment",
    "Bundle",
    "BundleSettingsValue",
    "Computation",
    "ComputationArgumentsValue",
    "ContourCv2",
    "CostFunction",
    "CreateApiKeyResponse",
    "CreateApiKeyrequest",
    "CreateUserRequest",
    "CreateUserResponse",
    "DictItem",
    "EquationExtraction",
    "EquationExtractionResponse",
    "EquationProcessingResponse",
    "EquationValidationResult",
    "ExecuteCodeResponse",
    "ExtractConstantsResponse",
    "ExtractTextResponse",
    "ExtractedPoint",
    "ExtractedPoints",
    "FindMappingResponse",
    "FindUserResponse",
    "FormalizeCircuitResponse",
    "FormalizeResponse",
    "GenerateCodeResponse",
    "GenerateComponentCodeResponse",
    "GenerateLensCodeResponse",
    "GetOptimizableParametersResponse",
    "GetSpectrumResponse",
    "GetSpectrumResponseSpectrumValueItem",
    "HttpValidationError",
    "InformalizeStatementResponse",
    "Net",
    "Netlist",
    "OptimizationHistory",
    "OptimizeConfig",
    "OptimizeNetlistResponse",
    "OptimizePlacementBodyResponse",
    "Parameter",
    "ParameterConstraint",
    "ParameterOutputLowerBound",
    "ParameterOutputUpperBound",
    "ParseResponse",
    "ParseStatementResponse",
    "PdkType",
    "PicInstance",
    "PicInstanceInfoValue",
    "PicInstanceSettingsValue",
    "PicWarning",
    "Placement",
    "PlotInfo",
    "PlotParserOutput",
    "RefineCodeResponse",
    "RefineComponentCodeResponse",
    "ScheduleJobResponse",
    "SolutionResponse",
    "SolutionResponseSolutionValue",
    "Statement",
    "StatementDictionary",
    "StatementType",
    "StatementValidation",
    "StatementValidationDictionary",
    "StatusResponse",
    "StructureConstraint",
    "StructureFunctionCall",
    "StructureFunctionCallArgumentsValue",
    "StructureFunctionCallExpectedResult",
    "SummarizerResponse",
    "ToolsListResponse",
    "UnformalizableStatement",
    "UnprocessableEntityError",
    "UserRequirement",
    "ValidateNetlistResponse",
    "ValidateResponse",
    "ValidationError",
    "ValidationErrorLocItem",
    "VariableRequirement",
    "VerifyCircuitCodeResponse",
    "VerifyResponse",
    "Z3Expression",
    "__version__",
    "code_execution",
    "document",
    "formalization",
    "fso",
    "lean",
    "pic",
    "requirements",
    "tools",
]



from IPython.display import display, Math, HTML  # type: ignore
import hypernetx as hnx  # type: ignore
import matplotlib.pyplot as plt  # type: ignore
import re


def display_full_results(validation_results, requirements=None, show_hypergraph=True):
    """Display equation validation results optimized for dark theme notebooks."""
    validations = validation_results.get("validations", {})

    matching = []
    non_matching = []

    for eq_name, value in validations.items():
        equation_data = {
            "name": eq_name,
            "latex": value.get("original_format", ""),
            "lhs": value.get("lhs_value"),
            "rhs": value.get("rhs_value"),
            "diff": abs(value.get("lhs_value", 0) - value.get("rhs_value", 0)),
            "percent_diff": abs(value.get("lhs_value", 0) - value.get("rhs_value", 0))
            / max(abs(value.get("rhs_value", 0)), 1e-10)
            * 100,
            "used_values": value.get("used_values", {}),
        }
        if value.get("is_valid"):
            matching.append(equation_data)
        else:
            non_matching.append(equation_data)

    # Summary header with dark theme
    total = len(validations)
    display(
        HTML(
            '<div style="background-color:#1e1e1e; padding:20px; border-radius:10px; margin:20px 0; '
            'border:1px solid #3e3e3e;">'
            f'<h2 style="font-family:Arial; color:#e0e0e0; margin-bottom:15px">Equation Validation Analysis</h2>'
            f'<p style="font-family:Arial; font-size:16px; color:#e0e0e0">'
            f"<b>Total equations analyzed:</b> {total}<br>"
            f'<span style="color:#4caf50">✅ Matching equations: {len(matching)}</span><br>'
            f'<span style="color:#ff5252">❌ Non-matching equations: {len(non_matching)}</span></p>'
            "</div>"
        )
    )

    # Non-matching equations
    if non_matching:
        display(
            HTML(
                '<div style="background-color:#2d1f1f; padding:20px; border-radius:10px; margin:20px 0; '
                'border:1px solid #4a2f2f;">'
                '<h3 style="color:#ff5252; font-family:Arial">⚠️ Equations Not Satisfied</h3>'
            )
        )

        for eq in non_matching:
            display(HTML(f'<h4 style="color:#e0e0e0; font-family:Arial">{eq["name"]}</h4>'))
            display(Math(eq["latex"]))
            display(
                HTML(
                    '<div style="font-family:monospace; margin-left:20px; margin-bottom:20px; '
                    "background-color:#2a2a2a; color:#e0e0e0; padding:15px; border-radius:5px; "
                    'border-left:4px solid #ff5252">'
                    f"Left side  = {eq['lhs']:.6g}<br>"
                    f"Right side = {eq['rhs']:.6g}<br>"
                    f"Absolute difference = {eq['diff']:.6g}<br>"
                    f"Relative difference = {eq['percent_diff']:.2f}%<br>"
                    "<br>Used values:<br>"
                    + "<br>".join([f"{k} = {v:.6g}" for k, v in eq["used_values"].items()])
                    + "</div>"
                )
            )

        display(HTML("</div>"))

    # Matching equations
    if matching:
        display(
            HTML(
                '<div style="background-color:#1f2d1f; padding:20px; border-radius:10px; margin:20px 0; '
                'border:1px solid #2f4a2f;">'
                '<h3 style="color:#4caf50; font-family:Arial">✅ Satisfied Equations</h3>'
            )
        )

        for eq in matching:
            display(HTML(f'<h4 style="color:#e0e0e0; font-family:Arial">{eq["name"]}</h4>'))
            display(Math(eq["latex"]))
            display(
                HTML(
                    '<div style="font-family:monospace; margin-left:20px; margin-bottom:20px; '
                    "background-color:#2a2a2a; color:#e0e0e0; padding:15px; border-radius:5px; "
                    'border-left:4px solid #4caf50">'
                    f"Value = {eq['lhs']:.6g}<br>"
                    "<br>Used values:<br>"
                    + "<br>".join([f"{k} = {v:.6g}" for k, v in eq["used_values"].items()])
                    + "</div>"
                )
            )

        display(HTML("</div>"))

    # Hypergraph visualization
    if show_hypergraph and requirements:
        display(
            HTML(
                '<div style="background-color:#1e1e1e; padding:20px; border-radius:10px; margin:20px 0; '
                'border:1px solid #3e3e3e;">'
                '<h3 style="color:#e0e0e0; font-family:Arial">🔍 Equation Relationship Analysis</h3>'
                '<p style="font-family:Arial; color:#e0e0e0">The following graph shows how variables are connected through equations:</p>'
                "</div>"
            )
        )

        list_api_requirements = requirements

        # Match get_eq_hypergraph settings exactly
        plt.rcParams["text.usetex"] = False
        plt.rcParams["mathtext.fontset"] = "stix"
        plt.rcParams["font.family"] = "serif"

        # Prepare hypergraph data
        hyperedges = {}
        for eq_name, details in validations.items():
            # Create a set of variables used in this equation
            used_vars = {f"${var}$" for var in details["used_values"].keys()}
            hyperedges[_get_latex_string_format(details["original_format"])] = used_vars

        # Create and plot the hypergraph
        H = hnx.Hypergraph(hyperedges)
        plt.figure(figsize=(16, 12))

        # Draw hypergraph with exact same settings as get_eq_hypergraph
        hnx.draw(
            H,
            with_edge_labels=True,
            edge_labels_on_edge=False,
            node_labels_kwargs={"fontsize": 14},
            edge_labels_kwargs={"fontsize": 14},
            layout_kwargs={"seed": 42, "scale": 2.5},
        )

        node_labels = list(H.nodes)
        symbol_explanations = _get_node_names_for_node_lables(node_labels, list_api_requirements)

        explanation_text = "\n".join([f"${symbol}$: {desc}" for symbol, desc in symbol_explanations])
        plt.annotate(
            explanation_text,
            xy=(1.05, 0.5),
            xycoords="axes fraction",
            fontsize=14,
            verticalalignment="center",
        )

        plt.title(r"Enhanced Hypergraph of Equations and Variables", fontsize=20)
        plt.show()

    return None


def _get_node_names_for_node_lables(node_labels, api_requirements):
    """
    Creates mapping between symbols and their descriptions.

    Args:
        node_labels: List of node labels (symbols) from the hypergraph
        api_requirements: Can be either:
            - List of dicts with {"latex_symbol": str, "requirement_name": str}
            - Dictionary mapping variable names to their descriptions
    """
    node_names = []

    # Handle case where api_requirements is a dictionary
    if isinstance(api_requirements, dict):
        for symbol in node_labels:
            clean_symbol = symbol.replace("$", "")
            if clean_symbol in api_requirements:
                node_names.append((clean_symbol, api_requirements[clean_symbol]))
        return node_names

    # Handle case where api_requirements is a list of dicts
    for symbol in node_labels:
        clean_symbol = symbol.replace("$", "")
        for req in api_requirements:
            if isinstance(req, dict) and req.get("latex_symbol") == clean_symbol:
                node_names.append((req["latex_symbol"], req["requirement_name"]))
                break

    return node_names


def _get_latex_string_format(input_string):
    """
    Properly formats LaTeX strings for matplotlib when text.usetex is False.
    No escaping needed since mathtext handles backslashes properly.
    """
    return f"${input_string}$"  # No backslash escaping required


def _get_requirements_set(requirements):
    variable_set = set()
    for req in requirements:
        variable_set.add(req["latex_symbol"])

    return variable_set


def _find_vars_in_eq(equation, variable_set):
    patterns = [re.escape(var) for var in variable_set]
    combined_pattern = r"|".join(patterns)
    matches = re.findall(combined_pattern, equation)
    return {rf"${match}$" for match in matches}
